# encoding: utf-8
# @Time   : 2021/8/4 12:01
# @author : zza
# @Email  : z740713651@outlook.com
# @File   : __init__.py

from simple_event_bus.core.async_event_bus import AsyncEventBus
from simple_event_bus.core.event import EVENT, EVENT_TYPE, Event
from simple_event_bus.core.event_bus import EventBus

__all__ = [Event, EVENT, EVENT_TYPE, AsyncEventBus, EventBus]



#!/usr/bin/python3
# encoding: utf-8
"""
A better experience in PyCharm:

Settings - Tools - Python integrated Tools - Testing - Default test runner: 'pytest'
"""

import os

import pytest


def setup_module(module):
    """setup any state specific to the execution of the given module."""
    pass


def teardown_module(module):
    """teardown any state that was previously setup with a setup_module method."""
    pass


if __name__ == "__main__":
    pytest.main(["-s", "--doctest-modules", os.path.dirname(os.path.dirname(__file__))])



# encoding: utf-8
# @Time   : 2021/8/4 12:04
# @author : zza
# @Email  : z740713651@outlook.com
# @File   : async_event_bus.py
import inspect
from typing import Callable, Union

from simple_event_bus.core.event import EVENT_TYPE, Event
from simple_event_bus.core.event_bus import EventBus


class AsyncEventBus(EventBus):
    async def _run_listener(
        self, event: Union[Event, EVENT_TYPE], listener: Callable
    ) -> Union[bool, None]:
        if inspect.iscoroutinefunction(listener):
            if self._need_event_param(listener):
                res = await listener(event)
            else:
                res = await listener()
        else:
            if self._need_event_param(listener):
                res = listener(event)
            else:
                res = listener()
        return res

    async def publish_event(self, event: Union[Event, EVENT_TYPE]) -> None:
        """Trigger the event list function.

        Both synchronous and asynchronous functions will trigger.
        if function return True, Other functions in the queue will not be triggered.

        Args:
            event: str, EVENT or instance of Event

        Returns:
            None
        """
        event = self.event_format(event)
        self.logger.debug(f"Get {event}")
        for listener in self._listeners[event.event_type]:
            res = await self._run_listener(event, listener)
            if res:
                # if listener return true. will break the loop
                self.logger.debug(f"{listener.__name__} break the loop")
                break



# encoding: utf-8
# @Time   : 2021/8/3 18:29
# @author : zza
# @Email  : z740713651@outlook.com
# @File   : errors.py


class EventBusBaseError(Exception):
    ...


class NotAsyncFunction(EventBusBaseError, TypeError):
    ...


class ErrorEventType(EventBusBaseError, TypeError):
    ...


class EVENTNameError(EventBusBaseError, NameError):
    """
    EVENT name can not contain spaces
    """


class MultiParamFunctionError(EventBusBaseError):
    """
    EventBus only send one param(The event) to function
    """


class DuplicateFunctionError(EventBusBaseError):
    """
    Duplicate Function in same EVENT
    """



#!/usr/bin/python3
# encoding: utf-8
from __future__ import print_function

from simple_event_bus.core import EVENT, EVENT_TYPE, AsyncEventBus, Event, EventBus
from simple_event_bus.event_source import (
    run_simple_event_source,
    run_simple_event_source_async,
)

from ._version import get_versions

__author__ = "AngusWG"
__email__ = "z740713651@outlook.com"
__version__ = get_versions()["version"]

del get_versions

__all__ = [
    __version__,
    # obj
    EVENT,
    EVENT_TYPE,  # for typing
    Event,
    EventBus,
    AsyncEventBus,
    # event_source
    run_simple_event_source,
    run_simple_event_source_async,
]



# encoding: utf-8
# @Time   : 2021/8/9 15:00
# @author : zza
# @Email  : z740713651@outlook.com
# @File   : __init__.py

import asyncio
import datetime
import time
from typing import Union

from simple_event_bus.core import EVENT, EVENT_TYPE, AsyncEventBus, Event, EventBus


class _LoopStatus:
    def __init__(self):
        self.loop_enable: bool = True


def run_simple_event_source(
    event_bus: EventBus,
    time_interval: Union[int, float] = 1,
    loop_event: EVENT_TYPE = EVENT("HeartBeat"),
    close_loop_event: EVENT_TYPE = EVENT("close_loop"),
) -> None:
    """
    A simple event source

    Args:
        event_bus: instance of EventBus.
        time_interval: loop gap interval(seconds).
        loop_event: the default loop event type.
        close_loop_event: close loop event type.

    Returns:
        None
    """
    loop_config = _LoopStatus()

    def close_loop(_: Event = None) -> None:
        """close loop"""
        loop_config.loop_enable = False

    event_bus.add_listener(close_loop_event, close_loop)
    loop_event = event_bus.event_type_format(loop_event)

    while loop_config.loop_enable:
        event_bus.publish_event(Event(loop_event, now=datetime.datetime.now()))
        time.sleep(time_interval)
    event_bus.remove_listener(close_loop_event, close_loop)
    return


async def run_simple_event_source_async(
    event_bus: AsyncEventBus,
    time_interval: Union[int, float] = 1,
    loop_event: EVENT_TYPE = EVENT("HeartBeat"),
    close_loop_event: EVENT_TYPE = EVENT("close_loop"),
) -> None:
    """
    A simple event source for async.

    Args:
        event_bus: instance of EventBus.
        time_interval: loop gap interval(seconds).
        loop_event: the default loop event type.
        close_loop_event: close loop event type.

    Returns:
        None
    """
    loop_config = _LoopStatus()

    def close_loop(_: Event = None) -> None:
        """close loop"""
        loop_config.loop_enable = False

    loop_config.loop_enable = True
    event_bus.add_listener(close_loop_event, close_loop)
    loop_event = event_bus.event_type_format(loop_event)

    while loop_config.loop_enable:
        await event_bus.publish_event(Event(loop_event, now=datetime.datetime.now()))
        await asyncio.sleep(time_interval)
    event_bus.remove_listener(close_loop_event, close_loop)
    return


__all__ = [run_simple_event_source, run_simple_event_source_async]



# encoding: utf-8
# @Time   : 2021/8/4 12:02
# @author : zza
# @Email  : z740713651@outlook.com
# @File   : event.py
from typing import TYPE_CHECKING, Any, Dict, Optional, TypeVar, Union

from simple_event_bus.errors import EVENTNameError

if TYPE_CHECKING:  # pragma: no cover
    # https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports
    from simple_event_bus.core import AsyncEventBus, EventBus


class EVENT(str):
    event_map: Dict[str, "EVENT"] = dict()

    def __new__(cls, name: str):
        if " " in name:
            raise EVENTNameError
        elif name in cls.event_map:
            return cls.event_map[name]
        else:
            item = super(EVENT, cls).__new__(cls, name)
            cls.event_map[name] = item
            return item


EVENT_TYPE = TypeVar("EVENT_TYPE", str, EVENT)


class Event(object):
    def __init__(self, event_type: EVENT_TYPE, **kwargs: Any):
        if isinstance(event_type, str):
            event_type = EVENT(event_type)
        self.__dict__: Dict[EVENT:Any] = kwargs
        self.event_type = event_type
        self.current_app: Optional[Union["EventBus", "AsyncEventBus"]] = None

    def __repr__(self) -> str:
        _attr = ", ".join(
            "{}={}".format(k, repr(v))
            for k, v in self.__dict__.items()
            if k != "event_type"
        )
        return f"Event({self.event_type}, {_attr})"

    def __getattribute__(self, *args, **kwargs) -> Any:  # for typing check
        return super(Event, self).__getattribute__(*args, **kwargs)



# encoding: utf-8
# @Time   : 2021/8/4 11:09
# @author : zza
# @Email  : z740713651@outlook.com
# @File   : test_core.py
import pytest

from simple_event_bus import run_simple_event_source, run_simple_event_source_async


class TestCore:
    def test_event(self):
        from simple_event_bus.core import EVENT
        from simple_event_bus.errors import EVENTNameError

        with pytest.raises(EVENTNameError):
            EVENT("a event")

    def test_basic_event_bus(self):
        from simple_event_bus.core import EVENT, Event, EventBus
        from simple_event_bus.errors import MultiParamFunctionError

        app = EventBus()
        event_list = []

        @app.listening("HeartBeat")
        def count(event: Event):
            event_list.append(event)
            if len(event_list) > 5:
                app.publish_event(Event(EVENT("close_loop")))
            return True

        run_simple_event_source(app, time_interval=0.1)
        assert len(event_list) > 5

        with pytest.raises(MultiParamFunctionError):

            @app.listening("HeartBeat")
            def error_func(a, b, c):
                print("this is a error function")
                return

    @pytest.mark.asyncio
    async def test_async_event_bus(self):
        from simple_event_bus.core import EVENT, AsyncEventBus, Event

        app = AsyncEventBus()

        event_list = []
        async_event_list = []

        @app.listening("HeartBeat")
        def count(event: Event):
            event_list.append(event)
            if len(event_list) <= 5:
                return True

        @app.listening(EVENT("HeartBeat"))
        async def async_count(event: Event):
            async_event_list.append(event)
            if len(async_event_list) == 5:
                await app.publish_event("close_loop")

        await run_simple_event_source_async(app, time_interval=0.1)
        assert len(event_list) == 10
        assert len(async_event_list) == 5
        assert "async_count" in app.get_listener_name_list("HeartBeat")
        assert "async_count" in app.get_listener_name_list()

    @pytest.mark.asyncio
    async def test_no_params_function(self):
        from simple_event_bus import AsyncEventBus, EventBus

        async_app = AsyncEventBus()

        @async_app.listening("HeartBeat")
        def func_1():
            return

        @async_app.listening("HeartBeat")
        async def func_2():
            return

        await async_app.publish_event("HeartBeat")

        sync_app = EventBus()
        sync_app.add_listener("HeartBeat", func_1)
        sync_app.publish_event("HeartBeat")

    def test_duplicate_function(self):
        from simple_event_bus import EventBus
        from simple_event_bus.errors import DuplicateFunctionError

        def func_1():
            return

        sync_app = EventBus()
        sync_app.add_listener("HeartBeat", func_1)
        with pytest.raises(DuplicateFunctionError):
            sync_app.add_listener("HeartBeat", func_1)

        sync_app.add_listener("HeartBeat", func_1, duplicate_check=False)

        sync_app.remove_listener("HeartBeat", func_1, iter_delete=True)



#!/usr/bin/python3
# encoding: utf-8

"""
test_simple_event_bus
----------------------------------

Tests for `simple_event_bus` module.
"""
import pytest


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get("https://github.com/audreyr/cookiecutter-pypackage")


class TestSimpleEventBus:
    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self):
        pass

    def teardown_method(self):
        pass



# encoding: utf-8
# @Time   : 2021/8/4 12:04
# @author : zza
# @Email  : z740713651@outlook.com
# @File   : event_bus.py
import functools
import inspect
import logging
from collections import defaultdict
from functools import lru_cache
from typing import Callable, Dict, List, Union

from simple_event_bus.core.event import EVENT, EVENT_TYPE, Event
from simple_event_bus.errors import DuplicateFunctionError, MultiParamFunctionError


class EventBus(object):
    def __init__(self):
        self._listeners: Dict[str, List[Callable]] = defaultdict(list)
        self.logger = logging.getLogger("simple_event_bus")

    @staticmethod
    @lru_cache()
    def event_type_format(event_type: EVENT_TYPE) -> EVENT:
        """Make str to EVENT"""
        if isinstance(event_type, str):
            return EVENT(event_type)

    @lru_cache()
    def event_format(self, event: Union[Event, EVENT_TYPE]) -> Event:
        """Make str or EVENT to be instance of Event"""
        if isinstance(event, str):
            event: EVENT = EVENT(event)
        if isinstance(event, EVENT):
            event: Event = Event(event)
        if event.current_app is None:
            event.current_app = self
        return event

    def add_listener(
        self, event_type: EVENT_TYPE, listener: Callable, duplicate_check: bool = True
    ) -> None:
        """Add function to event listener list.

        Args:
            event_type: str or instance of EVENT.
            listener: the function parameters must be (event: Event=None) or None.
            duplicate_check: if the function is already in list, raise DuplicateFunctionError

        Returns:
            None

        Examples:
            >>> EventBus().add_listener("heartbeat", lambda event:print(event))
        """
        # args check
        event_type = self.event_type_format(event_type)
        signature = inspect.signature(listener)
        if len(signature.parameters) > 1:
            raise MultiParamFunctionError(MultiParamFunctionError.__doc__)
        if duplicate_check and listener in self._listeners[event_type]:
            raise DuplicateFunctionError
        # append
        self._listeners[event_type].append(listener)

    def remove_listener(
        self, event_type: EVENT_TYPE, listener: Callable, iter_delete: bool = False
    ) -> None:
        """Remove function from listener list

        Args:
            event_type: str or instance of EVENT.
            listener: the function
            iter_delete: Multiple removals (if there are duplicate function at same list)

        Returns:
            None

        """
        # args check
        event_type = self.event_type_format(event_type)
        self._listeners[event_type].remove(listener)
        while iter_delete and listener in self._listeners[event_type]:
            self._listeners[event_type].remove(listener)

    def listening(self, event_type: EVENT_TYPE) -> Callable:
        """
        Add listener by wrapper

        Args:
            event_type: str or instance of EVENT.

        Examples:
            >>> app = EventBus()
            >>> @app.listening("HeartBeat")
            ... def heartbeat(event: Event=None):
            ...     return event.now
        """
        event_type = self.event_type_format(event_type)

        def wrapper(listener: Callable) -> Callable:
            self.add_listener(event_type, listener)
            return listener

        return wrapper

    @staticmethod
    @functools.lru_cache()
    def _need_event_param(listener: Callable) -> bool:
        """
        Determine if the event parameter is needed

        Run it every time. Cache it
        """
        signature = inspect.signature(listener)
        return len(signature.parameters) == 1

    def _run_listener(
        self, event: Union[Event, EVENT_TYPE], listener: Callable
    ) -> Union[bool, None]:
        if self._need_event_param(listener):
            res = listener(event)
        else:
            res = listener()
        return res

    def publish_event(self, event: Union[Event, EVENT_TYPE]) -> None:
        """Trigger the event list function

        if function return True, Other functions in the queue will not be triggered.

        Args:
            event: str, EVENT or instance of Event

        Returns:
            None
        """
        event = self.event_format(event)
        self.logger.debug(f"Get {event}")
        for listener in self._listeners[event.event_type]:
            res = self._run_listener(event, listener)
            if res:
                # if listener return true. will break the loop
                self.logger.debug(f"{listener.__name__} break the loop")
                break

    def get_listener_name_list(self, event_type: EVENT_TYPE = None) -> List[str]:
        if event_type is None:
            for listener_list in self._listeners.values():
                for listener in listener_list:
                    yield listener.__name__

        event_type = self.event_type_format(event_type)
        for listener in self._listeners[event_type]:
            yield listener.__name__



#!/usr/bin/python3
# encoding: utf-8
"""A setuptools based setup module for simple_event_bus"""

from codecs import open
from os import path
from setuptools import find_packages, setup

import versioneer

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open(path.join(here, "HISTORY.md"), encoding="utf-8") as history_file:
    history = history_file.read()

with open(path.join(here, "requirements.txt"), encoding="utf-8") as requirements_file:
    requirements = requirements_file.read()

with open(path.join(here, "requirements_dev.txt"), encoding="utf-8") as requirements_dev_file:
    requirements_dev = requirements_dev_file.read()

ext_modules = []
setup(
    name="simple_event_bus",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="a simple python event bus",
    long_description=readme + "\n\n" + history,
    long_description_content_type='text/markdown',
    author="AngusWG",
    author_email="z740713651@outlook.com",
    url="https://github.com/AngusWG/simple-event-bus",
    packages=find_packages(include=["simple_event_bus", "simple_event_bus.*"]),
    entry_points={
        "console_scripts": [
            # "simple_event_bus=simple_event_bus.__main__:entry_point",
        ],
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    extras_require={"dev_require": requirements + "\n" + requirements_dev},
    ext_modules=ext_modules,
    keywords="simple_event_bus",
)
